/********************************************************************* */
/* postprocess_column.c */
/* */
/* Postprocesses the disparity map using the column-wise method. */
/********************************************************************* */

/* Standard includes */
#include <assert.h>
#include <stdio.h>
#include <string.h>

/* Our includes */
#include "base.h"
#include "error.h"
#include "pnmio.h"
#include "p2p.h"

/* Special options for comparing our algorithm with other possibilities. */
/* For our algorithm, leave them all commented. */
/* #define HANDLE_NEARLY_HORIZONTAL_BORDERS */
/* #define ONLY_OVERWRITE_UNRELIABLE */


/* If printing of details during propagation is desired, set this variable */
/* to the column in which you are interested.  Otherwise, set it to a  */
/* negative value. */
static int col_interest = -1;
static int row_interest0 = 250, row_interest1 = 400;

/* Thresholds for reliability */
int th_reliable = 14;
double alpha = 0.15;
int th_moderately_reliable;
int th_slightly_reliable;

/* Threshold for max distance for aligning with gradient to handle  */
/* nearly horizontal boundaries */
int th_max_attraction = 10;

extern int writePostprocessingIntermediateResults;

static int verbose = 0;

#define abs(x) ((x) > 0 ? (x) : -(x))

/********************************************************************* */
/* setReliableThreshold */
/* setAlpha */
/* setMaxAttractionThreshold */
/* getReliableThreshold */
/* getAlpha */
/* getMaxAttractionThreshold */
/*  */
/* Provide an interface to the thresholds. */
/********************************************************************* */

void setReliableThreshold(int th)
{
  if (th < 0)  {
    warning("Reliable threshold must be nonnegative.  Setting to zero.");
    th = 0;
  }
  th_reliable = th;
}

void setAlpha(double a)
{
  if (a < 0)  {
    warning("Alpha must be nonnegative.  Setting to zero.");
    a = 0;
  }
  alpha = a;
}

void setMaxAttractionThreshold(int th)
{
  if (th < 0)  {
    warning("Max-attraction threshold must be nonnegative.  Setting to zero.");
    th = 0;
  }
  th_max_attraction = th;
}

int getReliableThreshold(void)
{
  return th_reliable;
}

double getAlpha(void)
{
  return alpha;
}

int getMaxAttractionThreshold(void)
{
  return th_max_attraction;
}

/********************************************************************* */
/* multiplay_uchar_mat_by_scalar */

void multiply_uchar_mat_by_scalar(
  uchar mat_in[ROWS][COLS],
  uchar mat_out[ROWS][COLS],
  int scalar)
{
  uchar *pi = (uchar *) mat_in;
  uchar *po = (uchar *) mat_out;
  int val;
  int i;
  
  for (i=ROWS*COLS ; i>0 ; i--) {
    val = *pi++ * scalar;
    val = max(val, 0);
    val = min(val, 255);
    *po++ = (uchar) val;
  }
}


void ceil_uint_mat(
  uint mat_in[ROWS][COLS],
  uchar mat_out[ROWS][COLS])
{
  uint *pi = (uint *) mat_in;
  uchar *po = (uchar *) mat_out;
  int val;
  int i;
  
  for (i=ROWS*COLS ; i>0 ; i--) {
    val = *pi++;
    val = max(val, 0);
    val = min(val, 255);
    *po++ = (uchar) val;
  }
}

void ceil_int_mat(
  int mat_in[ROWS][COLS],
  uchar mat_out[ROWS][COLS])
{
  int *pi = (int *) mat_in;
  uchar *po = (uchar *) mat_out;
  int val;
  int i;
  
  for (i=ROWS*COLS ; i>0 ; i--) {
    val = *pi++;
    val = max(val, 0);
    val = min(val, 255);
    *po++ = (uchar) val;
  }
}

void logical_or_mat(
  uchar mat1[ROWS][COLS],
  uchar mat2[ROWS][COLS],
  uchar mato[ROWS][COLS])
{
  uchar *pr1 = (uchar *) mat1;
  uchar *pr2 = (uchar *) mat2;
  uchar *pro = (uchar *) mato;
  int i;
  
  for (i=0 ; i<ROWS*COLS ; i++) {
    *pro = *pr1 || *pr2;
    pro++;
    pr1++;
    pr2++;
  }
}

/********************************************************************* */
/* WriteDMToPGM */
/* */
/* Writes a disparity map to a PGM file. */

void WriteDMToPGM(uchar dm[ROWS][COLS])
{
  uchar imgtmp[ROWS][COLS];
  char fname_out[80];
  static int count=0;
  
  multiply_uchar_mat_by_scalar(dm, imgtmp, 30);
  sprintf(fname_out, "output/dmt%04d.pgm", count);
  pgmWriteFile(fname_out, (uchar *) imgtmp, COLS, ROWS);
  count++;
}

void WriteRMToPGM(uint rm[ROWS][COLS])
{
  uchar imgtmp[ROWS][COLS];
  char fname_out[80];
  static int count=0;
  
  ceil_uint_mat(rm, imgtmp);
  sprintf(fname_out, "output/rmt%04d.pgm", count);
  pgmWriteFile(fname_out, (uchar *) imgtmp, COLS, ROWS);
  count++;
}


void WriteLabelsToPGM(int lab[ROWS][COLS])
{
  uchar imgtmp[ROWS][COLS];
  char fname_out[80];
  static int count=0;
  
  ceil_int_mat(lab, imgtmp);
  sprintf(fname_out, "output/labelst%04d.pgm", count);
  pgmWriteFile(fname_out, (uchar *) imgtmp, COLS, ROWS);
  count++;
}

/********************************************************************* */
/* compute_igxy */
/* */
/* Computes intensity gradients.  For each window of pixels, declares  */
/* the pixels within the window to be intensity gradients if the  */
/* intensity variation w/i the window is greater than a threshold. */
/* Interpretation:  igx = 1  means that pixel is an ig in the x-direction */
/*                  igy = 1     "      "      "       "       y-direction */
/********************************************************************* */

static void compute_igxy(uchar imgL[ROWS][COLS], 
                         uchar igx[ROWS][COLS], 
                         uchar igy[ROWS][COLS])
{
  int th = 5;                /* minimum variation w/i window */
  int w = 3;                 /* window width */
  int maxx, minn;
  int i, j, k;
  
  /* Initialize arrays */
  for (i = 0 ; i < ROWS ; i++)  {
    for (j = 0 ; j < COLS ; j++)  {
      igx[i][j] = 0;
      igy[i][j] = 0;
    }
  }
  
  /* Compute intensity gradients in x-direction */
  for (i = 0 ; i < ROWS ; i++)
    for (j = 0 ; j < COLS - w + 1 ; j++)  {
      maxx = 0;      minn = INF;
      for (k = j ; k < j + w ; k++)  {
        if (imgL[i][k] < minn)   minn = imgL[i][k];
        if (imgL[i][k] > maxx)   maxx = imgL[i][k];
      }
      if (maxx - minn >= th)
        for (k = j ; k < j + w ; k++)
          igx[i][k] = 1;
    }
  
  /* Compute intensity gradients in y-direction */
  for (j = 0 ; j < COLS ; j++)
    for (i = 0 ; i < ROWS - w + 1 ; i++)  {
      maxx = 0;      minn = INF;
      for (k = i ; k < i + w ; k++)  {
        if (imgL[k][j] < minn)   minn = imgL[k][j];
        if (imgL[k][j] > maxx)   maxx = imgL[k][j];
      }
      if (maxx - minn >= th)
        for (k = i ; k < i + w ; k++)
          igy[k][j] = 1;
    }
}


/********************************************************************* */
/* compute_igyd */
/*  */
/* Given igy (the boolean map of intensity gradients in the  */
/* y direction), computes igyd (the integer map of distances to ig's. */
/* Interpretation:  igyd = 0 means pixel is ig in y-direction */
/*                   "   = d means nearest ig is d pixels  below */
/*                   "   = -d   "     "     "       "      above */
/*                         (d is positive) */
/********************************************************************* */

static void compute_igyd(uchar igy[ROWS][COLS],
                         int igyd[ROWS][COLS])
{
  int i, j, curr_dist;
  
  for (j = 0 ; j < COLS ; j++)  {
    
    /* Expand downward */
    curr_dist = - 1 * INF;
    for (i = 0 ; i < ROWS ; i++)  {
      if (igy[i][j])  curr_dist = 0;
      else  curr_dist--;
      igyd[i][j] = curr_dist;
    }
    
    /* Expand upward */
    curr_dist = INF;
    for (i = ROWS - 1 ; i >= 0 ; i--)  {
      if (igy[i][j])  curr_dist = 0;
      else  curr_dist++;
      if (curr_dist < -1 * igyd[i][j]) igyd[i][j] = curr_dist;
    }
  }
  
}


/********************************************************************* */
/* compute_igxd */
/*  */
/* Given igx (the boolean map of intensity gradients in the  */
/* x direction), computes igxd (the integer map of distances to ig's). */
/* Interpretation:  igxd = 0 means pixel is ig in x-direction */
/*                   "   = d means nearest ig is d pixels to the right */
/*                   "   = -d   "     "     "       "     to the left  */
/*                         (d is positive) */
/********************************************************************* */

static void compute_igxd(uchar igx[ROWS][COLS],
                         int igxd[ROWS][COLS])
{
  int i, j, curr_dist;
  
  for (i = 0 ; i < ROWS ; i++)  {
    
    /* Expand rightward */
    curr_dist = - 1 * INF;
    for (j = 0 ; j < COLS ; j++)  {
      if (igx[i][j])  curr_dist = 0;
      else  curr_dist--;
      igxd[i][j] = curr_dist;
    }
    
    /* Expand leftward */
    curr_dist = INF;
    for (j = COLS - 1 ; j >= 0 ; j--)  {
      if (igx[i][j])  curr_dist = 0;
      else  curr_dist++;
      if (curr_dist < -1 * igxd[i][j]) igxd[i][j] = curr_dist;
    }
  }
}


/********************************************************************* */
/* computeReliabilitiesY */
/*  */
/* Labels each pixel with its "reliability".  Reliability of a pixel P */
/* is defined as the number of pixels in P's region, where the region is */
/* the contiguous set of pixels (in the y direction) with the same  */
/* disparity as P. */
/********************************************************************* */

static void computeReliabilitiesY(uchar disp_map[ROWS][COLS],
                                  uint reliability_map[ROWS][COLS])
{
  int i, j, k, curr_disp, curr_length;
  
  for (j = 0 ; j < COLS ; j++)  {
    curr_disp = -1;
    curr_length = 0;
    for (i = 0 ; i < ROWS ; i++)  {
      if (disp_map[i][j] == curr_disp)  curr_length++;
      else  {
        for (k = i - curr_length ; k < i ; k++)
          reliability_map[k][j] = curr_length;
        curr_disp = disp_map[i][j];
        curr_length = 1;
      }
    }
    for (k = i - curr_length ; k < i ; k++)
      reliability_map[k][j] = curr_length;
  }
}


/********************************************************************* */
/* computeReliabilitiesX */
/*  */
/* Same as computeReliabilitiesY, but in the x direction. */
/********************************************************************* */

static void computeReliabilitiesX(uchar disp_map[ROWS][COLS],
                                  uint reliability_map[ROWS][COLS])
{
  int i, j, k, curr_disp, curr_length;
  
  for (i = 0 ; i < ROWS ; i++)  {
    curr_disp = -1;
    curr_length = 0;
    for (j = 0 ; j < COLS ; j++)  {
      if (disp_map[i][j] == curr_disp)  curr_length++;
      else  {
        for (k = j - curr_length ; k < j ; k++)
          reliability_map[i][k] = curr_length;
        curr_disp = disp_map[i][j];
        curr_length = 1;
      }
    }
    for (k = j - curr_length ; k < j ; k++)
      reliability_map[i][k] = curr_length;
  }
}

/********************************************************************* */
/* propagateY */
/*  */
/* Propagates each moderately reliable region in the y direction until  */
/* it hits an intensity gradient.  (? First checks whether border of  */
/* reliable region is just beyond an intensity gradient.  If so, then  */
/* the region has probably extended into an object by accident, */
/* so therefore it backs up to align its border with the intensity  */
/* gradient.?)  Propagation occurs as long as neighboring region is  */
/* either unreliable or has a disparity at least two greater than the */
/* current region. */

static void propagateY(uchar disp_map[ROWS][COLS], 
                       uchar igy[ROWS][COLS],
                       int th_moderately_reliable, 
                       int th_slightly_reliable, 
                       uint reliability_map[ROWS][COLS],
                       int igyd[ROWS][COLS], 
                       int max_attraction)
{
  int i, j, k, curr_disp;
  
  for (j = 0 ; j < COLS ; j++)  {
    
    if (j==col_interest) {
      int oo;
      printf("Reliabilities of column %d\n", j);
      for (oo=row_interest0 ; oo<row_interest1 ; oo++) {
        printf("   Row %3d:  %3d\n", oo, reliability_map[oo][j]);
      }
      printf("\n");
    }
    
    i = 0;
    while (i < ROWS)  {
      
      /* Find top of new stable region */
      while (i < ROWS && reliability_map[i][j] < th_moderately_reliable)  i++;
      if (i >= ROWS)  break;
      curr_disp = disp_map[i][j];
      if (j == col_interest)  
        printf("top is i = %d, curr_disp = %d\n", i, curr_disp);
      
#ifdef HANDLE_NEARLY_HORIZONTAL_BORDERS
      /* Unless ig just below, ... */
      if (!(i > 0 && (reliability_map[i - 1][j] >= th_slightly_reliable)
          && igyd[i - 1][j] > 0 && igyd[i - 1][j] <= max_attraction))
#endif
      /* expand region upward */
        for (k = i - 1 ;
             k >= 0
               && !igy[k][j]
               && (reliability_map[k][j] < th_slightly_reliable 
                   || disp_map[k][j] == curr_disp
#ifndef ONLY_OVERWRITE_UNRELIABLE
                   || disp_map[k][j] >= curr_disp + 2
#endif
                   ) ;
             k-- )  {
          if (j == col_interest)  
            printf("prop up at [%d][%d] with disp = %d\n",  k, j, curr_disp);
          disp_map[k][j] = curr_disp;
          reliability_map[k][j] = th_moderately_reliable;
        }
    
      /* Find bottom of stable region */
      for (k = i + 1 ;
           k < ROWS && disp_map[k][j] == curr_disp ;
           k++) ;
      i = k - 1;
      if (k >= ROWS)  break;
    
      if (j == col_interest)  
        printf("bottom is i = %d, k = %d, curr_disp = %d\n", i, k, curr_disp);
    
#ifdef HANDLE_NEARLY_HORIZONTAL_BORDERS
      /* Unless ig just above, ... */
      if (i < ROWS - 1 && (reliability_map[i + 1][j] >= th_slightly_reliable)
          && igyd[i + 1][j] < 0 && -1 * igyd[i + 1][j] <= max_attraction) {
      }
      else
#endif
      {
        /* expand region downward */
        if (j==col_interest) printf("** k=%3d, curr_disp=%2d\n", k, curr_disp);
        
        for ( ;
              k < ROWS
                && !igy[k][j]
                && (reliability_map[k][j] < th_slightly_reliable 
                    || disp_map[k][j] == curr_disp
#ifndef ONLY_OVERWRITE_UNRELIABLE
                    || disp_map[k][j] >= curr_disp + 2
#endif
                    ) ;
              k++ )  {
          disp_map[k][j] = curr_disp;
          reliability_map[k][j] = th_moderately_reliable;
          if (j == col_interest) 
            printf("  propagating down into [%d][%d] with disp = %d\n", 
                   k, j, curr_disp);
        }
        i = k;
      } /* endelse */
    }  /* endwhile i */
  }  /* endfor j */
}
 

/********************************************************************* */
/* propagateX */
/*  */
/* Same as propagateY, except in the x direction. */

static void propagateX(uchar disp_map[ROWS][COLS], 
                       uchar igx[ROWS][COLS],
                       int th_moderately_reliable, 
                       int th_slightly_reliable, 
                       uint reliability_map[ROWS][COLS],
                       int igxd[ROWS][COLS],
                       int max_attraction)
{
  int i, j, k, curr_disp;

  for (i = 0 ; i < ROWS ; i++)  {

    if (i==col_interest) {
      int oo;
      printf("Reliabilities of row %d\n", i);
      for (oo=row_interest0 ; oo<row_interest1 ; oo++) {
        printf("   Column %3d:  %3d\n", oo, reliability_map[i][oo]);
      }
      printf("\n");
    }
    
    j = 0;
    while (j < COLS)  {

      /* Find left of new stable region */
      while (j < COLS && reliability_map[i][j] < th_moderately_reliable)  j++;
      if (j >= COLS)  break;
      curr_disp = disp_map[i][j];
      if (i == col_interest)  
         printf("left is j = %d, curr_disp = %d\n", j, curr_disp);

#ifdef HANDLE_NEARLY_HORIZONTAL_BORDERS
      /* Unless ig just to the right, ... */
      if (!(j > 0 && (reliability_map[i][j - 1] >= th_slightly_reliable)
          && igxd[i][j - 1] > 0 && igxd[i][j - 1] <= max_attraction))
#endif
        /* expand region leftward */
        for (k = j - 1 ;
             k >= 0
             && !igx[i][k]
             && (reliability_map[i][k] < th_slightly_reliable 
                 || disp_map[i][k] == curr_disp
#ifndef ONLY_OVERWRITE_UNRELIABLE
                 || disp_map[i][k] >= curr_disp + 2
#endif
                 ) ;
             k-- )  {
            if (i == col_interest)  
               printf("prop left at [%d][%d] with disp = %d\n",  i, k, curr_disp);
            disp_map[i][k] = curr_disp;
            reliability_map[i][k] = th_moderately_reliable;
        }

      /* Find right of stable region */
      for (k = j + 1 ;
           k < COLS && disp_map[i][k] == curr_disp ;
           k++) ;
      j = k - 1;
      if (k >= COLS)  break;

      if (i == col_interest)  
         printf("right is j = %d, k = %d, curr_disp = %d\n", j, k, curr_disp);

#ifdef HANDLE_NEARLY_HORIZONTAL_BORDERS
      /* Unless ig just to the left, ... */
      if (j < COLS - 1 && (reliability_map[i][j + 1] >= th_slightly_reliable)
          && igxd[i][j + 1] < 0 && -1 * igxd[i][j + 1] <= max_attraction) {
      }
      else
#endif
      {
        /* expand region rightward */
        for ( ;
              k < COLS
                && !igx[i][k]
                && (reliability_map[i][k] < th_slightly_reliable
                    || disp_map[i][k] == curr_disp
#ifndef ONLY_OVERWRITE_UNRELIABLE
                    || disp_map[i][k] >= curr_disp + 2
#endif
                    ) ;
              k++ )  {
          disp_map[i][k] = curr_disp;
          reliability_map[i][k] = th_moderately_reliable;
          if (i == col_interest) 
            printf("  propagating right into [%d][%d] with disp = %d\n", 
                   i, k, curr_disp);
        }
        j = k;
      }
    }  /* endwhile j */
  } /* endfor i */
}


/********************************************************************* */
/* removeIsolatedPixelsX */
/*  */
/* Given a binary map, removes isolated 1's in the x direction. */
/* */
/* INPUTS */
/* array:  binary array */
/* len:  smallest # of 1's that is allowed to remain. */
/*       For example, if len = 3, then */
/*          0000011000011100000 becomes */
/*          0000000000011100000. */
/* */
/* OUTPUTS */
/* array is changed */
/********************************************************************* */

static void removeIsolatedPixelsX(uchar array[ROWS][COLS], int len)
{
  int i, j, k, curr_len;

  for (i = 0 ; i < ROWS ; i++)  {
    curr_len = 0;
    for (j = 0 ; j < COLS ; j++)  {
      if (array[i][j])  curr_len++;
      else if (curr_len < len)  {
        for (k = j - curr_len ; k < j ; k++)
          array[i][k] = 0;
        curr_len = 0;
      }
      else  curr_len = 0;
    }  /* endfor j */
    if (curr_len < len)
      for (k = j - curr_len ; k < j ; k++)
        array[i][k] = 0;
  }  /* endfor i */
}


/********************************************************************* */
/* removeIsolatedPixelsY */
/*  */
/* Same as removeIsolatedPixelsX, except in the y direction. */
/********************************************************************* */

static void removeIsolatedPixelsY(uchar array[ROWS][COLS], int len)
{
  int i, j, k, curr_len;

  for (j = 0 ; j < COLS ; j++)  {
    curr_len = 0;
    for (i = 0 ; i < ROWS ; i++)  {
      if (array[i][j])  curr_len++;
      else if (curr_len < len)  {
        for (k = i - curr_len ; k < i ; k++)
          array[k][j] = 0;
        curr_len = 0;
      }
      else  curr_len = 0;
    }  /* endfor i */
    if (curr_len < len)
      for (k = i - curr_len ; k < i ; k++)
        array[k][j] = 0;
  }  /* endfor j */
}


/********************************************************************* */
/* coerceSurroundedPixelsY */
/*  */
/* Given an array, surrounded pixels are coerced.  In other words,  */
/* given a pixel P in the array, if P's two immediate neighbors in the  */
/* y direction (the one above and the one below) have the same value, */
/* then P's value is changed to agree with theirs. */
/********************************************************************* */

static void coerceSurroundedPixelsY(uchar array[ROWS][COLS])
{
  int i, j;

  for (i = 1 ; i < ROWS - 1 ; i++)
    for (j = 0 ; j < COLS ; j++)
      if (array[i-1][j] == array[i+1][j])
        array[i][j] = array[i-1][j];
}


/********************************************************************* */
/* modefilterY */
/*  */
/* Performs mode filtering in the y direction on an array whose  */
/* elements range from 0 to MAXDISP. */
/*  */
/* INPUTS */
/* array:  an array whose elements range from 0 to MAXDISP */
/* h:  filter height (must be odd) */
/*  */
/* NOTES */
/* This is an inefficient implementation, but it works. */
/********************************************************************* */

static void modefilterY(uchar array[ROWS][COLS], int h)
{
  int i, j, k, maxx, mode;
  int hh = h / 2;          /* filter half-height */
  int hist[MAXDISP + 1];   /* histogram */
  int disp;
  int inertia;

  if (h % 2 == 0)
    error("modefilterY:  filter height must be odd\n");

  for (j = 0 ; j < COLS ; j++)
    for (i = 0 ; i < ROWS ; i++)  {
      for (k = 0 ; k <= MAXDISP ; k++)
        hist[k] = 0;
      for (k = max(0,i-hh) ; k <= min(ROWS-1,i+hh) ; k++)  {
        disp = array[k][j];
        (hist[ disp ])++;   
      }

      disp = array[i][j];
      mode = disp;
      maxx = hist[disp];
/*      disp = array[i][j]; */
/*      maxx = hist[disp]; */
/*      if (disp>0) maxx += hist[disp-1]; */
/*      if (disp<MAXDISP) maxx += hist[disp+1]; */
/*      mode = disp; */
      for (k = 0 ; k <= MAXDISP ; k++)
        if (hist[k] > maxx)  {
          maxx = hist[k];
          mode = k;
        }
      inertia = hist[disp];
      if (disp>0) inertia += hist[disp-1];
      if (disp<MAXDISP) inertia += hist[disp+1];
      if (maxx>inertia) {
        array[i][j] = mode;
      } else {
        mode = disp;
        maxx = hist[disp];
        if (disp>0 && hist[disp-1]>maxx) {
          maxx = hist[disp-1];
          mode = disp-1;
        }
        if (disp<MAXDISP && hist[disp+1]>maxx) {
          mode = disp+1;
        }
        array[i][j] = mode;
      }
    }
}


/********************************************************************* */
/* modefilterX */
/*  */
/* Same as modefilterY, but in the x direction. */
/********************************************************************* */

static void modefilterX(uchar array[ROWS][COLS], int w)
{
  int i, j, k, maxx, mode;
  int hw = w / 2;          /* filter half-width */
  int hist[MAXDISP + 1];   /* histogram */
  int disp;
  int inertia;

  if (w % 2 == 0)
    error("modefilterX:  filter width must be odd\n");

  for (i = 0 ; i < ROWS ; i++)
    for (j = 0 ; j < COLS ; j++)  {
      for (k = 0 ; k <= MAXDISP ; k++)
        hist[k] = 0;
      for (k = max(0,j-hw) ; k <= min(COLS-1,j+hw) ; k++)  {
        disp = array[i][k];
        (hist[ disp ])++;
      }

      disp = array[i][j];
      mode = disp;
      maxx = hist[disp];
/*      disp = array[i][j]; */
/*      maxx = hist[disp]; */
/*      if (disp>0) maxx += hist[disp-1]; */
/*      if (disp<MAXDISP) maxx += hist[disp+1]; */
/*      mode = disp; */
      for (k = 0 ; k <= MAXDISP ; k++)
        if (hist[k] > maxx)  {
          maxx = hist[k];
          mode = k;
        }
      inertia = hist[disp];
      if (disp>0) inertia += hist[disp-1];
      if (disp<MAXDISP) inertia += hist[disp+1];
      if (abs(mode-disp)<=1 || maxx>inertia) {
        array[i][j] = mode;
      } else {
        mode = disp;
        maxx = hist[disp];
        if (disp>0 && hist[disp-1]>maxx) {
          maxx = hist[disp-1];
          mode = disp-1;
        }
        if (disp<MAXDISP && hist[disp+1]>maxx) {
          mode = disp+1;
        }
        array[i][j] = mode;
      }
    }
}


/********************************************************************* */
/* computeDepthDiscontinuities */
/*  */
/* Given a disparity map, computes the depth discontinuities as those  */
/* pixels that border a change in disparity of at least two levels,  */
/* and that lie on the background.  (This latter condition is necessary */
/* to prevent two neighboring pixels from both being declared  */
/* discontinuities.) */
/********************************************************************* */

static void computeDepthDiscontinuities(uchar disp_map[ROWS][COLS], 
                                        uchar dd_map[ROWS][COLS])
{
  int i, j;

  for (i = 1 ; i < ROWS - 1 ; i++)
    for (j = 1 ; j < COLS - 1 ; j++)  {
      if (   disp_map[i][j] < disp_map[i+1][j] - 1
          || disp_map[i][j] < disp_map[i-1][j] - 1
          || disp_map[i][j] < disp_map[i][j+1] - 1
          || disp_map[i][j] < disp_map[i][j-1] - 1 )
        dd_map[i][j] = DISCONTINUITY;
      else dd_map[i][j] = NO_DISCONTINUITY;
    }

  for (i = 0 ; i < ROWS ; i++)  {
    dd_map[i][0] = NO_DISCONTINUITY;
    dd_map[i][COLS-1] = NO_DISCONTINUITY;
  }
  for (j = 0 ; j < COLS ; j++)  {
    dd_map[0][j] = NO_DISCONTINUITY;
    dd_map[ROWS-1][j] = NO_DISCONTINUITY;
  }
}


/********************************************************************* */
/* postprocess */
/* */
/* Version used in STAN-CS-TR-96-1573 and ICCV '98 */
/* */
/* Postprocess the disparity map.  The basic idea is to propagate */
/* reliable regions into unreliable regions, stopping at intensity  */
/* gradients.  (The actual rules are more complicated, of course.)   */
/* Further processing, like mode filtering, helps to clean up things. */
/*  */
/* INPUTS */
/* imgL:  original left intensity image */
/* dm_orig:  disparity map computed by matching the scanlines independently */
/*  */
/* OUTPUTS */
/* disp_map:  disparity map after postprocessing */
/* dd_map:  depth discontinuities after postprocessing */
/********************************************************************* */

void postprocess(uchar imgL[ROWS][COLS], 
                 uchar imgR[ROWS][COLS],   /* unused in this routine */
                 uchar dm_orig[ROWS][COLS], 
                 uchar disp_map[ROWS][COLS], 
                 uchar dd_map[ROWS][COLS])
{
  uint reliability_map[ROWS][COLS];       /* reliabilities of pixels */
  uchar igx[ROWS][COLS], igy[ROWS][COLS]; /* intensity gradients */
  int igyd[ROWS][COLS], igxd[ROWS][COLS]; /* distance to nearest gradient */

  th_slightly_reliable = (int) (th_reliable*(1-alpha) + 0.5);
  th_moderately_reliable = (int) (th_reliable*(1+alpha) + 0.5);

  printf("Parameters:  sr=%d, mr=%d",
         th_slightly_reliable, th_moderately_reliable);
#ifdef HANDLE_NEARLY_HORIZONTAL_BORDERS
  printf(", ma=%d\n", th_max_attraction);
#else
  printf("\n");
#endif

  if ((th_slightly_reliable > th_moderately_reliable))
    error("postprocess:  Reliability thresholds do not obey monotonicity.");
  if (row_interest0 < 0 || row_interest0 >= ROWS ||
      row_interest1 < 0 || row_interest1 >= ROWS)
    row_interest1 = row_interest0 - 1;

  /* Copy original disparity map to new disparity map */
  memcpy((uchar *) disp_map, (uchar *) dm_orig, ROWS*COLS);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("dm0.pgm", (uchar *) disp_map, COLS, ROWS);

  /* Remove "obvious errors" in the disparity map */
  coerceSurroundedPixelsY(disp_map);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("dm1.pgm", (uchar *) disp_map, COLS, ROWS);

  /* Compute intensity gradients */
  compute_igxy(imgL, igx, igy);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("igx0.pgm", (uchar *) igx, COLS, ROWS);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("igy0.pgm", (uchar *) igy, COLS, ROWS);

  /* Remove "isolated" intensity gradients in the y direction */
  removeIsolatedPixelsX(igy, 3);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("igy1.pgm", (uchar *) igy, COLS, ROWS);

  /* Propagate reliable regions in the y direction */
  compute_igyd(igy, igyd);
  computeReliabilitiesY(disp_map, reliability_map);
  propagateY(disp_map, igy, th_moderately_reliable, th_slightly_reliable,
             reliability_map, igyd, th_max_attraction);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("dm2.pgm", (uchar *) disp_map, COLS, ROWS);

  /* Remove "isolated" intensity gradients in the x direction */
  removeIsolatedPixelsY(igx, 3);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("igx1.pgm", (uchar *) igx, COLS, ROWS);

  /* Propagate reliable regions in the x direction */
  compute_igxd(igx, igxd);
  computeReliabilitiesX(disp_map, reliability_map);
  propagateX(disp_map, igx, th_moderately_reliable, th_slightly_reliable,
             reliability_map, igxd, th_max_attraction);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("dm3.pgm", (uchar *) disp_map, COLS, ROWS);

  /* Mode filter the disparity map */
  modefilterY(disp_map, 11);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("dm4.pgm", (uchar *) disp_map, COLS, ROWS);
  modefilterX(disp_map, 11);
  if (writePostprocessingIntermediateResults)
	pgmWriteFile("dm5.pgm", (uchar *) disp_map, COLS, ROWS);

  /* Find the depth discontinuities from the disparity map */
  computeDepthDiscontinuities(disp_map, dd_map);
}



